from .exceptions import InvalidPenalizerConfigError
from .penalizer_interface import PenalizerInterface
from .late_penalizer import LatePenalizer
from .checkoff_penalizer import CheckoffPenalizer

__all__ = ['InvalidPenalizerConfigError', 'PenalizerInterface',
           'LatePenalizer', 'CheckoffPenalizer']

_PENALIZERS = (
    LatePenalizer,
    CheckoffPenalizer
)
AVAILABLE_PENALIZERS = {cls.__name__: cls for cls in _PENALIZERS}



from .exceptions import InvalidGraderConfigError
from .grader_interface import GraderInterface, Part
from .threaded_grader import ThreadedGrader
from .prompt_grader import PromptGrader
from .open_file_grader import OpenFileGrader
from .command_grader import CommandGrader
from .libcheck_grader import LibcheckGrader
from .junit_json_grader import JUnitJSONGrader
from .junit_xml_grader import JUnitXMLGrader
from .bitwise_json_grader import BitwiseJSONGrader
from .circuitsim_grader import CircuitSimGrader
from .pylc3_grader import PyLC3Grader
from .multi_command_grader import MultiCommandGrader

__all__ = ['InvalidGraderConfigError', 'GraderInterface', 'Part',
           'ThreadedGrader', 'PromptGrader', 'OpenFileGrader', 'CommandGrader',
           'LibcheckGrader', 'JUnitJSONGrader', 'JUnitXMLGrader',
           'BitwiseJSONGrader', 'CircuitSimGrader', 'PyLC3Grader',
           'MultiCommandGrader']

_GRADERS = (
    PromptGrader,
    OpenFileGrader,
    CommandGrader,
    LibcheckGrader,
    JUnitJSONGrader,
    JUnitXMLGrader,
    BitwiseJSONGrader,
    CircuitSimGrader,
    PyLC3Grader,
    MultiCommandGrader,
)
AVAILABLE_GRADERS = {cls.__name__: cls for cls in _GRADERS}



"""Flatten a submission like the old SubmissionFix script"""

import os
import tarfile
import posixpath
import shutil
import zipfile
from abc import ABCMeta, abstractmethod

from .utils import mkdir_p, sanitize_path

# XXX Slow because unlike Marie's SubmissionFix script, does not use
#     native extractors
# XXX Does not support symlinks properly. Ignores them in tar archives
#     and writes out their destination in zip archives
# XXX Does not support empty directories
# XXX No error handling


class ArchiveError(Exception):
    """Raised for corrupt archives"""
    pass


class Archive:
    """Handle an archive format for the extraction code"""

    __metaclass__ = ABCMeta

    def __init__(self, path):
        self.path = path

    def __enter__(self):
        self.open()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def path(self):
        """Return the path to this archive"""
        return self.path

    @abstractmethod
    def open(self):
        """
        Open the archive at self.path and store it somewhere like
        self.archive.
        """
        pass

    @abstractmethod
    def close(self):
        """Close archive"""
        pass

    @abstractmethod
    def uncompressed_size(self):
        """
        Return the total size of all the files in the archive in bytes.
        """
        pass

    @abstractmethod
    def names(self):
        """Return a list of filenames in this archive."""
        pass

    @abstractmethod
    def file(self, name):
        """Return a file-like object pointing to name"""
        pass


class TarArchive(Archive):
    def open(self):
        try:
            self.archive = tarfile.open(self.path)
        except tarfile.TarError as err:
            raise ArchiveError(err)

    def close(self):
        try:
            self.archive.close()
        except tarfile.TarError as err:
            raise ArchiveError(err)

    def uncompressed_size(self):
        try:
            return sum(member.size for member in self.archive.getmembers())
        except tarfile.TarError as err:
            raise ArchiveError(err)

    def names(self):
        try:
            return [member.name for member in self.archive.getmembers()
                    if member.isfile()]
        except tarfile.TarError as err:
            raise ArchiveError(err)

    def file(self, name):
        try:
            return self.archive.extractfile(name)
        except tarfile.TarError as err:
            raise ArchiveError(err)


class ZipArchive(Archive):
    def open(self):
        try:
            self.archive = zipfile.ZipFile(self.path)
        except zipfile.BadZipfile as err:
            raise ArchiveError(err)

    def close(self):
        try:
            self.archive.close()
        except zipfile.BadZipfile as err:
            raise ArchiveError(err)

    def uncompressed_size(self):
        try:
            return sum(info.file_size for info in self.archive.infolist())
        except zipfile.BadZipfile as err:
            raise ArchiveError(err)

    def names(self):
        try:
            return [info.filename for info in self.archive.infolist()
                    if info.filename and info.filename[-1] != '/']
        except zipfile.BadZipfile as err:
            raise ArchiveError(err)

    def file(self, name):
        try:
            return self.archive.open(name)
        except zipfile.BadZipfile as err:
            raise ArchiveError(err)


# To protect against zipbombs, refuse to extract if archive contents are
# larger than 64 MiB. This may need adjusting if students try to upload
# 300 MiB pictures of Diddy for their gba games
MAX_UNCOMPRESSED_SIZE_BYTES = 64 * 2**20

ARCHIVE_TYPES = {
    '.tgz': TarArchive,
    '.tar.gz': TarArchive,
    '.tar.bz2': TarArchive,
    '.tbz2': TarArchive,
    '.tar.xz': TarArchive,
    '.txz': TarArchive,
    '.zip': ZipArchive,
}


def extract(archive, dest_dir, max_archive_size=None):
    """
    Extract Archive to X, flattening it as needed
    """

    if max_archive_size is None:
        max_archive_size = MAX_UNCOMPRESSED_SIZE_BYTES

    with archive:
        # Don't do zipbomb detection if the user switches it off by
        # passing a max size <= 0
        if max_archive_size > 0:
            uncompressed_size = archive.uncompressed_size()
            if uncompressed_size > max_archive_size:
                # In case this error bubbles its way up to students,
                # don't reveal the setup of their filesystem
                safe_archive_name = os.path.basename(archive.path())
                raise ArchiveError("Archive `{}' has uncompressed size of {} "
                                   "bytes which exceeds maximum of {} bytes. "
                                   "Refusing to extract!"
                                   .format(safe_archive_name,
                                           uncompressed_size,
                                           max_archive_size))

        names = {}

        for name in archive.names():
            # Remove backslash memes from malformed zip files
            safe_name = name.replace('\\', '/')
            components = sanitize_path(safe_name, path_lib=posixpath,
                                       join=False)

            if '__MACOSX' in components:
                # Skip macOS "Resource Forks"
                continue
            elif not components:
                # Empty filename
                continue

            names[name] = components

        if not names:
            # Empty archive!
            return

        prefix = next(iter(names.values()))

        # Try to find a prefix
        while prefix:
            for component in names.values():
                if len(prefix) >= len(component) \
                        or component[:len(prefix)] != prefix:
                    break
            else:
                # Cool, we found the prefix!
                break

            prefix = prefix[:-1]

        # Strip prefix
        if prefix:
            for name in names:
                names[name] = names[name][len(prefix):]

        # Extract files
        for name in names:
            dirname = os.path.join(dest_dir, *names[name][:-1])
            basename = names[name][-1]

            mkdir_p(dirname)
            with open(os.path.join(dirname, basename), 'wb') as extracted, \
                    archive.file(name) as archived:
                shutil.copyfileobj(archived, extracted)


def flatten(files_dir, max_archive_size=None):
    """
    Search the top-level files in `files_dir' for archives, and extract
    each one. Check for zipbombs.
    """

    for file_ in os.listdir(files_dir):
        for extension, archive_cls in ARCHIVE_TYPES.items():
            if file_.lower().endswith(extension):
                archive_path = os.path.join(files_dir, file_)
                extract(archive_cls(archive_path), files_dir, max_archive_size)
                # There's not much use wasting disk space by keeping
                # the archive around now that we've extracted it, so
                # delete it.
                os.remove(archive_path)
                break



import os
import yaml
import shutil
from zipfile import ZipFile

from ..utils import sanitize_path, mkdir_p

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader


class GradescopeLoader(object):
    def __init__(self, zipfile_path):
        self.zipfile_path = zipfile_path
        self.zipfile = None
        self.rootdir = None
        self.submissions = {}

    def __enter__(self):
        self.zipfile = ZipFile(self.zipfile_path)
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.zipfile.close()

    def load(self):
        # Gradescope was designed by Bay Area folks
        self.rootdir = self.zipfile.namelist()[0].split('/', maxsplit=1)[0]
        with self.zipfile.open('{}/submission_metadata.yml'
                               .format(self.rootdir)) as metadata:
            meta = yaml.load(metadata, Loader=SafeLoader)

        self.submissions = {submission_id: data[':submitters'][0][':name']
                            for submission_id, data in meta.items()}

    def extract_files(self, submission_id, dest_dir):
        prefix = '{}/{}/'.format(self.rootdir, submission_id)

        for name in self.zipfile.namelist():
            if not name.startswith(prefix) or name.endswith('/'):
                continue

            basename = sanitize_path(name[len(prefix):])
            fullpath = os.path.join(dest_dir, basename)
            mkdir_p(os.path.dirname(fullpath))

            with self.zipfile.open(name) as sourcefile, \
                    open(fullpath, 'wb') as destfile:
                shutil.copyfileobj(sourcefile, destfile)



import json
import os
from fractions import Fraction

from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, STDOUT, TimeoutExpired
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework using pyLC3 tests
"""


class PyLC3Test(Part):
    __slots__ = ('test',)

    def __init__(self, test):
        self.test = test

    def description(self):
        return self.test

    def grade(self, result):
        if result is None:
            return PartGrade(score=Fraction(0),
                             log='results for test not found. misspelling?')

        log = '\n'.join('{0[display-name]}: {0[message]}'.format(test)
                        for test in result if not test['passed'])
        failed = sum(1 for test in result if not test['passed'])

        score = Fraction(len(result) - failed, len(result))
        return PartGrade(score=score, log=log)


class PyLC3Grader(GraderInterface):
    """
    Run a pyLC3 grader and collect the results.
    """

    DEFAULT_TIMEOUT = 30

    def __init__(self, test_class, timeout=None):
        self.test_class = test_class

        if timeout is None:
            self.timeout = self.DEFAULT_TIMEOUT
        else:
            self.timeout = timeout

    def list_prerequisites(self):
        return ['build-essential', 'g++', 'cmake', 'libboost-all-dev',
                'libglib2.0-dev', 'castxml', 'python-pip']

    def list_extra_setup_commands(self):
        return ['pip uninstall -y wheel',  # THANKS ARJUN
                'pip install scikit-build',
                'pip install pyLC3',
                'ldconfig',
                'pip install parameterized']

    def part_from_config_dict(self, config_dict):
        return PyLC3Test.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        cmdline = ['python2', 'runner.py', self.test_class]
        # env['LD_LIBRARY_PATH'] =
        # '/usr/local/lib/python2.7/dist-packages/pyLC3'
        try:
            # Do not mix stderr into stdout because sometimes our friend
            # Roi printStackTrace()s or System.err.println()s, and that
            # will mess up JSON parsing
            process = run_process(cmdline, cwd=path, timeout=self.timeout,
                                  stdout=PIPE, stderr=STDOUT, input='')
        except TimeoutExpired:
            raise BrokenSubmissionError('timeout of {} seconds expired for '
                                        'grader'.format(self.timeout))

        if process.returncode != 0:
            raise BrokenSubmissionError(
                'grader command exited with nonzero exit code {}'
                .format(process.returncode),
                verbose=(process.stdout.decode()
                         if process.stdout else '(no output)'))

        with open(os.path.join(path, 'results.json'), 'r') as json_fp:
            results = json.load(json_fp)

        return [part.grade(results['results'].get(part.test))
                for part in parts]



from abc import ABCMeta, abstractmethod

from ..utils import ConfigDictMixin


class GraderInterface(ConfigDictMixin):
    __metaclass__ = ABCMeta

    def __init__(self):
        """
        The class needs an init method that will take in all of its
        desired options from the config file as keyword args. Required
        options that need to be in the config file should not have
        default values. Optional ones can have default values but still
        need to exist as kwargs. Each of the elements in the options:
        section of the component config of the assignment will be passed
        as a keyword argument during initialization.
        """
        pass

    def list_prerequisites(self):
        """
        This function should return a list of Ubuntu 16.04 packages
        required to run this grader.
        """
        return []

    def list_extra_setup_commands(self):
        """
        This function should return a list of extra one-time commands to
        run at Docker image creation time. This is Ubuntu.
        """
        return []

    def is_interactive(self):
        """
        Return True if and only if this grader will produce command-line
        prompts.
        """
        return False

    def needs_display(self):
        """
        Return True if and only if this grader expects a graphical
        environment, like $DISPLAY on GNU/Linux. Does not necessarily
        imply is_interactive() is True since some graders are
        noninteractive but still connect to a display server, like
        CircuitSim graders (since CircuitSim needs JavaFX).
        """
        return False

    @abstractmethod
    def part_from_config_dict(self, config_dict):
        """
        Convert and validate a dictionary parsed from the `parts'
        section of a component configuration to a Part instance.
        """
        pass

    @abstractmethod
    def grade(self,
              submission,
              path,
              parts):
        """
        This function should take in a Submission object and a path,
        where the path can be assumed to be the root of the submission
        directory (it will be a directory where the grading manager has
        copied the required files for this grader); then complete the
        grading on it, and then return a list of kasjdfkasjdfkj
        instances containing the result for each subcomponent.
        """
        pass


class Part(ConfigDictMixin):
    """
    Represent a `part' of grading which has its own weight and its own
    score: one prompt question, one unit test, etc.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        """
        Subclasses should override this method and include
        required/optional config keys as arguments. Then you can simply
        return MyPartSubclass.from_config_dict(config_dict) in
        MyGrader.part_from_config_dict().
        """
        pass

    @abstractmethod
    def description(self):
        """
        Return a human-friendly description for this part. Used in grade
        breakdowns and logs.
        """
        pass



import threading
from multiprocessing import cpu_count
from abc import ABCMeta, abstractmethod

from . import GraderInterface
from ..utils import queue


"""
An implementation of GraderInterface which runs tests in separate
threads. In theory, speeds up grading by two times the number of
CPUs.

Actual graders can subclass this and implement grade_part() to become
threaded!
"""


class ThreadedGrader(GraderInterface):
    __metaclass__ = ABCMeta

    """
    A base class for graders to run parts in separate threads to speed
    up grading. Subclasses must implement grade_part().
    """

    def __init__(self, num_threads=None):
        if num_threads is None:
            try:
                cpus = cpu_count()
            except NotImplementedError:
                cpus = 1

            self.num_threads = 2 * cpus
        else:
            self.num_threads = num_threads

    @abstractmethod
    def grade_part(self, part, path, submission):
        """
        Grade a Part instance part, where the temporary grading
        directory is `path' and the Submission instance passed to
        grade() is `submission'.

        Must return a PartGrade instance.
        """
        pass

    def run_thread(self, path, submission, part_queue, grades):
        """
        One grading thread. Dequeues parts from part_queue until
        part_queue is empty and exits.
        """

        while True:
            try:
                part_index, part = part_queue.get(block=False)
            except queue.Empty:
                return

            try:
                grade = self.grade_part(part, path, submission)
            except Exception as err:
                grades[part_index] = err
                return

            grades[part_index] = grade

    def grade(self, submission, path, parts):
        """Spin off the configured number of grading threads and grade."""

        grades = [None] * len(parts)
        threads = []
        part_queue = queue.Queue()

        for index_part in enumerate(parts):
            part_queue.put(index_part)

        for _ in range(self.num_threads):
            args = (path, submission, part_queue, grades)
            thread = threading.Thread(target=self.run_thread, args=args)
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

        # Look for exceptions
        for grade in grades:
            if isinstance(grade, Exception):
                raise grade

        return grades



import shlex

from . import PromptGrader, InvalidGraderConfigError
from ..utils import run_process


class CommandGrader(PromptGrader):
    def __init__(self, command):
        # Set up the Prompt Grader first
        super(CommandGrader, self).__init__()

        if not isinstance(command, str) or not command:
            raise InvalidGraderConfigError(
                "A command needs to be specified.")
        else:
            self.command = shlex.split(command)

    def grade(self, submission, path, parts):
        # Let's get the file from the submission
        run_process(self.command, cwd=path)
        return super(CommandGrader, self).grade(submission=submission,
                                                path=path, parts=parts)



import os
import shutil
import tempfile
from fractions import Fraction
from collections import namedtuple

import click
import yaml

from .submission import BrokenSubmissionError
from .grades import AssignmentComponentGrade, CalculatedGrade, \
                    CalculatedPenalty
from .graders import AVAILABLE_GRADERS
from .penalizers import AVAILABLE_PENALIZERS
from .constants import ASSIGNMENT_CONFIG_FILE, ASSIGNMENT_FILES_DIRECTORY
from .utils import ConfigDictMixin, copy_globs, sanitize_path


class ComponentPart(namedtuple('ComponentPart',
                               ['weight', 'part', 'partial_credit'])):
    def calculate_grade(self, component_points, total_part_weight, part_grade):
        points = component_points * Fraction(self.weight, total_part_weight)
        return part_grade.calculate_grade(points,
                                          self.part,
                                          self.partial_credit)


class AssignmentComponent(ConfigDictMixin):
    def __init__(self, assignment, name, backend, weight, parts, files=None,
                 optional_files=None, grading_files=None,
                 backend_options=None):
        self.assignment = assignment
        self.name = name
        self.backend = backend

        # Get the backend class
        if self.backend not in AVAILABLE_GRADERS:
            raise ValueError("Invalid grading backend: %s" % backend)

        backend_class = AVAILABLE_GRADERS[self.backend]

        self.weight = weight
        if type(self.weight) != int:
            raise ValueError("Component weights need to be integers.")

        self.files = files
        if self.files is not None:
            if not isinstance(self.files, list):
                raise ValueError('List of files must be a list')
            else:
                self.files = [sanitize_path(file_) for file_ in self.files]

        self.optional_files = optional_files
        if self.optional_files is not None:
            if not isinstance(self.optional_files, list):
                raise ValueError('List of optional files must be a list')
            else:
                self.optional_files = [sanitize_path(file_)
                                       for file_ in self.optional_files]

        # Check that there are no files marked as both optional and
        # required
        if None not in (self.files, self.optional_files):
            common = set(self.files).intersection(self.optional_files)
            if common:
                raise ValueError(
                    "file `{}' cannot be both an optional and required file!"
                    .format(next(iter(common))))

        # TODO: Confirm that all of the files in the grading list exist
        self.grading_files = grading_files
        if self.grading_files is not None:
            if not isinstance(self.grading_files, list):
                raise ValueError('List of grading files must be a list')
            else:
                self.grading_files = [sanitize_path(file_)
                                      for file_ in self.grading_files]

        # We then initialize the grader
        self.grader = backend_class.from_config_dict(backend_options or {})
        self.parts = []
        self.total_part_weight = 0

        for part_dict in parts:
            if 'weight' not in part_dict:
                raise ValueError('every part needs a weight!')
            else:
                weight = part_dict['weight']
                del part_dict['weight']

            partial_credit = True
            if 'partial-credit' in part_dict:
                partial_credit = part_dict['partial-credit']
                del part_dict['partial-credit']

            part = self.grader.part_from_config_dict(part_dict)
            self.parts.append(ComponentPart(weight=weight,
                                            part=part,
                                            partial_credit=partial_credit))
            self.total_part_weight += weight

    def is_interactive(self):
        """
        Return True if and only if this component will produce
        command-line prompts.
        """
        return self.grader.is_interactive()

    def list_prerequisites(self):
        """
        Return a list of Ubuntu 16.04 packages required to run this
        grader.
        """
        return self.grader.list_prerequisites()

    def list_extra_setup_commands(self):
        """
        This function should return a list of extra one-time commands to
        run at Docker image creation time. This is Ubuntu.
        """
        return self.grader.list_extra_setup_commands()

    def needs_display(self):
        """
        Return True if and only if this grader expects a graphical
        environment, like $DISPLAY on GNU/Linux.
        """
        return self.grader.needs_display()

    def grade_submission(self, submission):
        grading_directory = tempfile.mkdtemp(prefix='zucchini-component-')

        try:
            # Copy the submission first and the grading later so that if a file
            # exists in both, the grading copy overwrites the submission copy
            if self.optional_files:
                submission.copy_files(self.optional_files, grading_directory,
                                      allow_fail=True)
            if self.files:
                submission.copy_files(self.files, grading_directory)
            if self.grading_files:
                self.assignment.copy_files(self.grading_files,
                                           grading_directory)
            # self.parts is a list of (weight, Part) tuples, but we only
            # wanna pass Part instances to the grader
            parts = [part.part for part in self.parts]
            part_grades = self.grader.grade(submission, grading_directory,
                                            parts)
        except BrokenSubmissionError as err:
            return AssignmentComponentGrade(error=str(err),
                                            error_verbose=err.verbose)
        finally:
            shutil.rmtree(grading_directory)

        return AssignmentComponentGrade(part_grades)

    def calculate_grade(self, total_weight, component_grade):
        points = Fraction(self.weight, total_weight)
        return component_grade.calculate_grade(points, self.name,
                                               self.total_part_weight,
                                               self.parts)


class AssignmentPenalty(ConfigDictMixin):
    """Penalize students for late submissions etc."""

    def __init__(self, assignment, name, backend, backend_options=None):
        self.assignment = assignment
        self.name = name

        # Get the backend class
        if backend not in AVAILABLE_PENALIZERS:
            raise ValueError("Invalid penalizer backend: %s" % backend)

        backend_class = AVAILABLE_PENALIZERS[backend]

        # We then initialize the grader
        self.penalizer = backend_class.from_config_dict(backend_options or {})

    def calculate(self, submission, grade):
        """Return `grade' as a Fraction adjusted for the given submission"""

        adjusted_grade = self.penalizer.adjust_grade(submission, grade)
        return CalculatedPenalty(name=self.name,
                                 points_delta=adjusted_grade - grade)


# This class contains the Assignment configuration for the local file
class Assignment(object):
    def __init__(self, root):
        self.root = root

        config_file_path = os.path.join(self.root, ASSIGNMENT_CONFIG_FILE)

        if not os.path.exists(config_file_path):
            raise ValueError("This directory is not a valid Zucchini "
                             "assignment: the Zucchini config is missing.")

        with click.open_file(config_file_path, 'r') as config_file:
            config = yaml.safe_load(config_file)

        if config is None:
            raise ValueError("The assignment configuration file could not be "
                             "parsed.")

        try:
            # Get the assignment's name, author
            self.name = config['name']
            self.author = config['author']
        except KeyError as e:
            raise ValueError("Missing field in assignment config: %s" %
                             e.args[0])

        # TODO: Don't hardcode Canvas logic in here. Need to make something
        #       like "Modules" for handling these things.
        if 'canvas' in config:
            try:
                self.canvas_course_id = int(config['canvas']['course-id'])
                self.canvas_assignment_id = \
                    int(config['canvas']['assignment-id'])
            except KeyError as err:
                raise ValueError('canvas section in assignment config is '
                                 'missing key: {}'.format(err.args[0]))
            except ValueError as err:
                raise ValueError('canvas id is not an integer: {}'
                                 .format(str(err)))
        else:
            self.canvas_course_id = None
            self.canvas_assignment_id = None

        self.components = []

        # Now load the components (this is the fun part!)
        for component_config in config['components']:
            component = AssignmentComponent.from_config_dict(component_config,
                                                             assignment=self)
            self.components.append(component)

        if len({component.name for component in self.components}) \
                != len(self.components):
            raise ValueError('Duplicate component names')

        self.total_weight = sum(c.weight for c in self.components)
        self.interactive = any(c.is_interactive() for c in self.components)
        self.noninteractive = any(not c.is_interactive()
                                  for c in self.components)
        self._needs_display = any(c.needs_display() for c in self.components)

        self.prerequisites = set()
        for component in self.components:
            self.prerequisites.update(component.list_prerequisites())

        self.extra_setup_commands = []
        backend_types_seen = set()
        for component in self.components:
            if component.backend not in backend_types_seen:
                backend_types_seen.add(component.backend)
                self.extra_setup_commands += \
                    component.list_extra_setup_commands()

        self.penalties = []

        for penalty_config in config.get('penalties', ()):
            penalty = AssignmentPenalty.from_config_dict(penalty_config,
                                                         assignment=self)
            self.penalties.append(penalty)

        # TODO: Handle assignments with no components or with 0 total weight

    def list_prerequisites(self):
        """
        Return a list of Ubuntu 16.04 packages needed by this assignment.
        """
        return list(self.prerequisites)

    def list_extra_setup_commands(self):
        """
        Return a list of Ubuntu commands to run at setup time.
        """
        return self.extra_setup_commands

    def has_interactive(self):
        """
        Return True if and only if grading this assignment will produce
        command-line prompts.
        """
        return self.interactive

    def has_noninteractive(self):
        """
        Return True if and only if grading this assignment contains at
        least one noninteractive grader.
        """
        return self.noninteractive

    def needs_display(self):
        """
        Return True if and only if at least one component expects a
        graphical environment, like $DISPLAY on GNU/Linux.
        """
        return self._needs_display

    def copy_files(self, files, path):  # (List[str], str) -> None
        """Copy the grader files in the files list to the new path"""

        grading_files_dir = os.path.join(self.root, ASSIGNMENT_FILES_DIRECTORY)
        # XXX Replace FileNotFoundError raised with a better exception
        copy_globs(files, grading_files_dir, path)

    def grade_submission(self, submission, interactive=None):
        """
        Grade each assignment component of submission, returning an
        AssignmentComponentGrade for each component.
        If interactive is None (the default), grade all components; if
        True, grade only interactive components, and if False, grade
        only noninteractive components.
        """

        # The grading data for each part (individual test) of each
        # component (test suite) of this assignment
        grades = [component.grade_submission(submission)
                  if interactive is None
                  or interactive == component.is_interactive() else None
                  for component in self.components]
        return grades
        # TODO: We probably want to log, too

    def calculate_grade(self, submission, component_grades):
        """
        Calculate the final grade for a submission given the list of
        AssignmentComponentGrades for the submission.
        """

        # First, calculate raw grade
        grade = CalculatedGrade(name=self.name, grade=Fraction(1),
                                raw_grade=Fraction(0), penalties=[],
                                components=[])

        for component, component_grade in zip(self.components,
                                              component_grades):
            calc_component_grade = component.calculate_grade(self.total_weight,
                                                             component_grade)
            grade.components.append(calc_component_grade)
            grade.grade += calc_component_grade.points_delta

        # Store grade pre-penalties
        grade.raw_grade = grade.grade

        # Now deduct for penalties
        for penalty in self.penalties:
            calc_penalty = penalty.calculate(submission, grade.grade)
            grade.penalties.append(calc_penalty)
            grade.grade += calc_penalty.points_delta

        return grade



"""
Utilities for gradescope autograding.
"""

import os
import json
from fractions import Fraction
from zipfile import ZipFile, ZIP_DEFLATED

from . import __version__ as ZUCCHINI_VERSION
from .constants import ASSIGNMENT_CONFIG_FILE, ASSIGNMENT_FILES_DIRECTORY
from .utils import ConfigDictMixin, ConfigDictNoMangleMixin, \
                   datetime_from_string, recursive_get_using_string


class GradescopeMetadata(object):
    """
    Parse the metadata as described in:
    https://gradescope-autograders.readthedocs.io/en/latest/submission_metadata/
    """

    _ATTRS = [
        ('student_name', 'users.0.name', str),
        ('submission_date', 'created_at', datetime_from_string),
        ('due_date', 'assignment.due_date', datetime_from_string),
        # The nested int(float(..)) deal is because int('100.0')
        # explodes
        ('total_points', 'assignment.outline.0.weight',
            lambda pts: int(float(pts))),
    ]

    def __init__(self, json_dict):
        for attr, key, type_ in self._ATTRS:
            val = recursive_get_using_string(json_dict, key)
            setattr(self, attr, type_(val))

    @classmethod
    def from_json_path(cls, json_path):
        with open(json_path, 'r') as json_fp:
            return cls(json.load(json_fp))


class GradescopeAutograderTestOutput(ConfigDictNoMangleMixin, ConfigDictMixin):
    """
    Output of a single test in Gradescope JSON.
    """

    def __init__(self, name=None, score=None, max_score=None, output=None):
        self.name = name
        self.score = score
        self.max_score = max_score
        self.output = output


class GradescopeAutograderOutput(ConfigDictNoMangleMixin, ConfigDictMixin):
    """
    Hold Gradescope Autograder output as described in
    https://gradescope-autograders.readthedocs.io/en/latest/specs/#output-format
    """

    def __init__(self, score=None, tests=None, extra_data=None):
        self.score = score
        self.tests = [GradescopeAutograderTestOutput.from_config_dict(test)
                      for test in tests] if tests is not None else None
        self.extra_data = extra_data

    def to_config_dict(self, *args):
        dict_ = super(GradescopeAutograderOutput, self).to_config_dict(*args)
        if dict_.get('tests', None):
            dict_['tests'] = [test.to_config_dict() for test in dict_['tests']]
        return dict_

    @staticmethod
    def _two_decimals(grade, frac):
        """Convert a fraction to string with two decimal points"""
        return '{:.02f}'.format(grade.to_float(frac))

    @classmethod
    def from_grade(cls, grade):
        """
        Convert a grading_manager.Grade to Gradescope JSON.
        """

        score = grade.score()
        tests = []
        # Store the component grades in the extra_data field
        extra_data = {'component_grades': grade.serialized_component_grades()}

        computed_grade = grade.computed_grade()

        # Add penalties
        for penalty in computed_grade.penalties:
            if penalty.points_delta != 0:
                # Hack: Display -37 as 0/37 and +37 as 37/37
                fake_max_score = cls._two_decimals(
                    grade, abs(penalty.points_delta))
                fake_score = cls._two_decimals(grade, Fraction(0)) \
                    if penalty.points_delta < 0 else fake_max_score
                test = GradescopeAutograderTestOutput(
                    name=penalty.name,
                    score=fake_score,
                    max_score=fake_max_score)
                tests.append(test)

        # Add actual test results
        for component in computed_grade.components:
            if component.error:
                test = GradescopeAutograderTestOutput(
                    name=component.name,
                    score=cls._two_decimals(grade, component.points_got),
                    max_score=cls._two_decimals(
                        grade, component.points_possible),
                    output='{}\n{}'.format(component.error,
                                           component.error_verbose or ''))
                tests.append(test)
            else:
                for part in component.parts:
                    if part.deductions:
                        deductions = 'Deductions: {}\n\n'.format(
                            ', '.join(part.deductions))
                    else:
                        deductions = ''

                    test = GradescopeAutograderTestOutput(
                        name='{}: {}'.format(component.name, part.name),
                        score=cls._two_decimals(grade, part.points_got),
                        max_score=cls._two_decimals(
                            grade, part.points_possible),
                        output=deductions + part.log)
                    tests.append(test)

        return cls(score=score, tests=tests, extra_data=extra_data)

    def to_json_stream(self, fp):
        json.dump(self.to_config_dict(), fp)


SETUP_SH = r'''#!/bin/bash
# THIS FILE WAS GENERATED BY ZUCCHINI
set -e

cd /autograder/source
# Prevent apt from prompting for input and hanging the build
export DEBIAN_FRONTEND=noninteractive
apt-get update
apt-get install -y python3 python3-pip python3-wheel {prereqs}
pip3 install {pip_install_arg}
{extra_setup_commands}
'''


RUN_AUTOGRADER = r'''#!/bin/bash
# THIS FILE WAS GENERATED BY ZUCCHINI
set -e
set -o pipefail

cd /autograder/source
zucc flatten /autograder/submission
{grade_cmd_prefix}zucc grade-submission /autograder/submission \
    | zucc gradescope bridge /autograder/submission_metadata.json \
    > /autograder/results/results.json
'''


RUN_GRAPHICAL_SH = r'''#!/bin/bash

cat >xorg.conf <<'EOF'
# This xorg configuration file is meant to be used by xpra
# to start a dummy X11 server.
# For details, please see:
# https://xpra.org/Xdummy.html

Section "ServerFlags"
  Option "DontVTSwitch" "true"
  Option "AllowMouseOpenFail" "true"
  Option "PciForceNone" "true"
  Option "AutoEnableDevices" "false"
  Option "AutoAddDevices" "false"
EndSection

Section "Device"
  Identifier "dummy_videocard"
  Driver "dummy"
  Option "ConstantDPI" "true"
  VideoRam 192000
EndSection

Section "Monitor"
  Identifier "dummy_monitor"
  HorizSync   5.0 - 1000.0
  VertRefresh 5.0 - 200.0
  Modeline "1024x768" 18.71 1024 1056 1120 1152 768 786 789 807
EndSection

Section "Screen"
  Identifier "dummy_screen"
  Device "dummy_videocard"
  Monitor "dummy_monitor"
  DefaultDepth 24
  SubSection "Display"
    Viewport 0 0
    Depth 24
    Modes "1024x768"
    Virtual 1024 768
  EndSubSection
EndSection
EOF

/usr/lib/xorg/Xorg -noreset -logfile ./xorg.log -config ./xorg.conf :69 \
    >/dev/null 2>&1 &

xorg_pid=$!

export DISPLAY=:69
"$@"

exitcode=$?

kill "$xorg_pid" || {
    printf 'did not kill Xorg!\n' >&2
    exit 1
}

exit $exitcode'''


class GradescopeAutograderZip(object):
    """
    Generates a Gradesope autograder zip file from which Gradescope
    generates a Docker image for grading.
    """

    def __init__(self, path='.', prerequisites=None, extra_setup_commands=None,
                 needs_display=False, wheel_path=None):
        self.path = path
        self.prerequisites = prerequisites or []
        self.extra_setup_commands = extra_setup_commands or []
        self.needs_display = needs_display
        self.wheel_path = wheel_path

        # Need this for
        if self.needs_display:
            prerequisites.append('xserver-xorg-video-dummy')

    def _relative_path(self, abspath):
        """
        Convert an absolute path to an assignment file to a path
        relative to self.path.
        """
        return os.path.relpath(abspath, self.path)

    def _real_path(self, relpath):
        """
        Convert a relative path to an assignment file to an absolute
        path.
        """
        return os.path.join(self.path, relpath)

    def _write_file(self, file_path, zipfile, real_path=None):
        """
        Add a file to the generated zip file. file_path is the
        destination path in the .zip file. If real_path is not provided,
        it will be self.path/file_path.
        """
        if real_path is None:
            real_path = self._real_path(file_path)
        zipfile.write(real_path, file_path)

    def _write_string(self, string, path, zipfile):
        """
        Add a file to the generated zip file. file_path should be relative to
        self.path.
        """
        zipfile.writestr(path, string)

    def _write_dir(self, dir_path, zipfile):
        """
        Recursively add a directory to the generated zip file. dir_path
        should be relative to self.path.
        """

        real_path = self._real_path(dir_path)

        for dirpath, _, filenames in os.walk(real_path):
            for filename in filenames:
                relpath = self._relative_path(os.path.join(dirpath, filename))
                self._write_file(relpath, zipfile)

    def write_zip(self, file):
        """
        Write the autograder .zip to file. If file is a file-like
        object, write it there, otherwise it should be a string
        designating the destination path.
        """

        with ZipFile(file, 'w', ZIP_DEFLATED) as zipfile:
            self._write_file(ASSIGNMENT_CONFIG_FILE, zipfile)

            grading_files = self._real_path(ASSIGNMENT_FILES_DIRECTORY)
            if os.path.exists(grading_files):
                self._write_dir(ASSIGNMENT_FILES_DIRECTORY, zipfile)

            if self.needs_display:
                self._write_string(RUN_GRAPHICAL_SH, 'run_graphical.sh',
                                   zipfile)
                grade_cmd_prefix = 'bash run_graphical.sh '
            else:
                grade_cmd_prefix = ''

            run_autograder = RUN_AUTOGRADER.format(
                grade_cmd_prefix=grade_cmd_prefix)
            self._write_string(run_autograder, 'run_autograder', zipfile)

            if self.wheel_path is None:
                pip_install_arg = 'zucchini==' + ZUCCHINI_VERSION
            else:
                # Can't just name it `zucchini.whl' or something because
                # this upsets pip
                wheel_filename = os.path.basename(self.wheel_path)
                self._write_file(wheel_filename, zipfile,
                                 real_path=self.wheel_path)
                pip_install_arg = wheel_filename

            extra_setup_commands = '\n'.join(self.extra_setup_commands)
            setup_sh = SETUP_SH.format(
                pip_install_arg=pip_install_arg,
                prereqs=' '.join(self.prerequisites),
                extra_setup_commands=extra_setup_commands)
            self._write_string(setup_sh, 'setup.sh', zipfile)
